import numpy as np
import statistics as st
import scipy.linalg as sl

import math
from matplotlib import pyplot as plt
import pandas as pd
from itertools import combinations

########## Diversity mesures ##########

# average distance between all objects
def structdist( distrib, distance ):
    card = np.sum( distrib )
    result = 0
    i = 0
    for cat in distrib:
        j = 0
        for ocat in distrib:
            if i != j:
                result += cat*ocat*distance[i][j]
            j = j+1
        i = i+1
    result = result/(card*(card-1))
    # compute standard deviation
    stdev = 0
    i = 0
    for cat in distrib:
        j = 0
        for ocat in distrib:
            stdev += cat*ocat*((result-distance[i][j])**2)
            j = j+1
        i = i+1
    stdev = math.sqrt(stdev/(card*(card)))
    return result,stdev

# diversity-based-on entropy, the naïve way
def entropy( distrib, distance, q ):
    card = np.sum(distrib)
    result = 0
    i = 0
    for cat in distrib:
        if cat != 0:
            j = 0
            intresult = 0
            for ocat in distrib:
                if ocat != 0:
                    intresult += (ocat/card)*math.exp(-distance[i][j])
                j = j+1
            #print(intresult)
            #print(q)
            result += (cat/card)*math.pow(intresult,q-1)
        i = i+1
    #print(result)
    #print(q)
    result = math.pow(result,1/(1-q))
    return result

# normalised entropy (the very naïve way)
# The rational is that the entropy is maximal when all categories are evenly filled
# and it is scale free (does not depend on their size)
# However, this is not true as demonstrated by the example of graphsemdist with the curves below...
# Hence we compare with the unitary distribution
def nentropy( distrib, distance, q ):
    full = []
    for cat in distrib:
        full.append( 1 )
    max = entropy( full, distance, q )
    return (entropy( distrib, distance, q )-1) / (max-1)

# A vectorized form of the conversion from distance to similarity
# Beware: it MUST be normalized (see Leinster 2012)
def yasserdistsim(m):
    max = np.max(m)
    return 1-(m/max)
#print( semstructdist1 )
#print( distsim( semstructdist1 ) )

def distsim(m):
    return np.exp(-m)

def relativeabundance( distrib ):
    card = np.sum(distrib) # cardinal of the distribution
    norm = lambda t: t/card
    return norm( distrib ) # relative abundance
#print( distribs[3] )
#print( relativeabundance( distribs[3] ) )

# Main body of the generalised diversity-entropy function
def summation( distrib, sim, exp ):
    #print( "Summation for "+str(distrib)+" with q="+str(q)+" and "+str(sim) )
    rab = relativeabundance( distrib ) # relative abundance
    #print( rab )
    result = 0
    i = 0
    for cat in distrib:
        if cat != 0:
            j = 0
            intresult = 0
            for ocat in distrib:
                if ocat != 0:
                    intresult += rab[j]*sim[i][j]
                j = j+1
            result += rab[i]*math.pow(intresult,exp)
        i = i+1
    return result

# Specific version for q=1
def prodation( distrib, sim ):
    rab = relativeabundance( distrib ) # relative abundance
    result = 1
    i = 0
    for cat in distrib:
        if cat != 0:
            j = 0
            intresult = 0
            for ocat in distrib:
                if ocat != 0:
                    intresult += rab[j]*sim[i][j]
                j = j+1
            result *= math.pow(intresult,-rab[i])
        i = i+1
    return result

# Recomputed using the functions before (more efficient, same results)
def diversity( distrib, distance, q ):
    if q == 1:
        return prodation( distrib, distsim( distance ) )
    else:
        sum = summation( distrib, distsim( distance ), q-1 )
        #print( math.pow(sum,1/(1-q)) )
        return math.pow(sum,1/(1-q))

# See nentropy
def ndiversity( distrib, distance, q ):
    full = []
    for cat in distrib:
        full.append( 1 )
    max = diversity( full, distance, q )
    return ( diversity( distrib, distance, q )-1) / (max-1)
#print( diversity( distribs[8], graphsemdist2, 2 ) )
#print( ndiversity( distribs[8], graphsemdist2, 2 ) )
#print( entropy( distribs[3], graphsemdist2, 2 ) )
#print( nentropy( distribs[3], graphsemdist2, 2 ) )
#print( diversity( distribs[3], unstructdist, -2 ) )
#print( ndiversity( distribs[3], unstructdist, -2 ) )
#print( entropy( distribs[3], unstructdist, -2 ) )
#print( nentropy( distribs[3], unstructdist, -2 ) )

# diameter: largest distance between two non zero elements
def calcdiam( distrib, distance ):
    largest = 0
    i = 0
    # I could assume it symmetric!
    for cat in distrib:
        if cat > 0:
            j = 0
            for ocat in distrib:
                if ocat > 0:
                    if distance[i,j] > largest:
                        largest = distance[i,j]
                j = j+1
        i = i+1
    return largest

# median computed the classical way
def median( distrib, distance ):
    # get the distance distribution and its cardinal
    listvals = []
    card = 0
    i = 0
    for cat in distrib:
        if cat > 0:
            j = 0
            for ocat in distrib:
                if ocat > 0: # add 'i != j and' for suppressing distance 0
                    listvals += [[distance[i,j],cat*ocat]]
                    card += cat*ocat
                j = j+1
        i = i+1
    # add one, divide by two, take the indeger value
    cut = int((card+1)/2)
    # sort the set of distance
    listvals.sort( key=lambda x: x[0] )
    # get the two intermediate values
    v1 = 0
    v2 = 0
    for [ dist, quant ] in listvals:
        cut -= quant
        if cut <= 0:
            if v1 == 0:
                v1 = dist
                if cut < 0:
                    v2 = dist
                    break
            else:
                v2 = dist
                break
    # average them
    return (v1+v2)/2
#print ( median( [1, 2, 3], np.array([[0,.5,.5],[.5,0,.5],[.5,.5,0]]) ) )
#print ( median( [1,1,1,1,1], np.array([[0,.1,.1,.1.9],[.1,0,.1,.1,.9],[.9,.9,0]]) ) )

# A simple way to measure Shanon entropy (not the associated diversity)
def shannon( distrib ):
    card = np.sum(distrib)
    result = 0
    i = 0
    for cat in distrib:
        if distrib[i] > 0:
            result += (distrib[i]/card)*math.log(distrib[i]/card,2)
        i = i+1
    return -result

########## Establishing the diversity of various distributions ##########

# Returns the best diversity in one operation   
def oneop( distrib, div ):
    initval = div( distrib )
    best = list(distrib)
    val = -1
    i = 0
    for cat in distrib:
        if cat != 0:
            j = 0
            for ocat in distrib:
                if ocat != 0 and i != j:
                    current = list(distrib)
                    current[i] += 1
                    current[j] -= 1
                    #print(current)
                    curdiv = div( current )
                    #print(curdiv)
                    if curdiv < initval and curdiv > val:
                        best = current
                        val = curdiv
                j += 1
        i += 1
    return best

def walkdistribs( distrib, div ):
    distlist = [ [ distrib, div( distrib )]]
    dist = distrib
    while True:
        newdist = oneop( dist, div )
        if newdist == dist:
            break
        distlist.append( [ newdist, div( newdist )] )
        dist = newdist
    return distlist

def selectdistribs( distrib, distance, num ):
    distlist = walkdistribs( distrib, distance )
    #print( distlist )
    # I know that this is not very pythonic, better leave the user read the stacktrace
    if len( distlist ) < num:
        print( "Cannot find that many distributions: "+str(num)+" out of "+str(len( distlist )) )
        return
    maxdiv = distlist[0][1]
    mindiv = distlist[-1][1]
    chunk = (maxdiv - mindiv) / (num-1)
    #print( chunk )
    index = [ distlist[0] ]
    num -= 1
    curtarget = maxdiv - chunk
    i = 1
    while num > 1:
        while distlist[i][1] > curtarget:
            i += 1
        if (distlist[i-1][1] - curtarget) < (curtarget - distlist[i][1]):
            i -= 1
        num -= 1
        index.append( distlist[i] )
        curtarget -= chunk
        i += 1
    index.append( distlist[-1] )
    return index 

# Renders the list of value as a real distribution
def distrib_as_list( distribution, base ):
    res = []
    j=0
    for i in range(max(base) + 1):
        if i not in base:
            res += [0.]
        else:
            res += [distribution[j]]
            j += 1
    return res

########## Finding the maximizing distribution for an ontology ##########

# Return distribution max of a Z matrix
# Z is n*n

#Remember to change similarity to distance with distsim(matrix) use this function correctly
def find_max_dist(Z):
    n = Z.shape[0]
    Dmax = 0
    subsetBMax = None
    weight = None
    Zb = None
    distribution = None

    for k in range(1, n + 1):
        for B in combinations(range(n), k):
            submatrix = Z[np.ix_(B, B)]

            #Dmax, subsetBMax, weight, Zb, distribution = find_max_distrib_flat( submatrix )
            try:
                # solve the linear equation to find weight of Zb
                w = np.linalg.solve(submatrix, np.ones(len(B))) 
                if np.all(w >= 0): #weight has to be non negative
                    magnitude = np.sum(w) 
                    if magnitude > Dmax: #if current magnitude is > dMax, then it become Dmax
                        Dmax = magnitude
                        subsetBMax = B
                        weight = w
                        Zb = submatrix
                        distribution = np.divide(weight,Dmax)
                    if magnitude == Dmax: #take the case where magnitude == Dmax so that the biggest matrix is the choosen one
                        subsetBMax = B
                        weight = w
                        Zb = submatrix
                        distribution = np.divide(weight,Dmax)

            except np.linalg.LinAlgError:
                # exception for singular matrix 
                pass

    return Dmax, subsetBMax, weight, Zb, distribution

def find_max_distrib_flat(Z):
    weights = np.linalg.solve( Z, np.ones(Z.shape[0]))
    #if np.all( weights >= 0): #weight has to be non negative
    Dmax = np.sum(weights)
    return Dmax, list( np.arange( 0, Z.shape[0], 1) ), weights, Z, np.divide(weights,Dmax)

# This function can be used if the similarity matrix has no 0 on the diagonal
# It is garantee to converge if it is strictly diagonal dominant or symmetric and positive defined
def find_max_dist_gs(Z):
    x = gauss_seidel(Z)
    subsetBMax = list(np.where(x >= 0.)[0])
    
    weight = x[subsetBMax]
    Zb = Z[np.ix_(subsetBMax, subsetBMax)]
    distribution = np.divide(weight, np.sum(weight))
    distribution = distribution[distribution > 1e-3]
    Dmax = np.sum(weight)
    
    return Dmax, tuple(subsetBMax), weight, Zb, distribution

# This function can be used if the similarity matrix is semi-definite positive
def find_max_dist_dp(Z):
    return find_max_distrib_flat(Z)

########## Gauss-Seidel optimisation ##########

# This is a copy of the code from https://primer-computational-mathematics.github.io/book/c_mathematics/numerical_methods/15_iterative_methods_to_solve_matrix.html
# The only modifications are the parameters
def gauss_seidel(Z):
    maxit=500
    tol=1.e-4
    b = np.ones(len( Z ))
    m, n = Z.shape
    x = np.zeros(Z.shape[0])
    residuals = []
    for k in range(maxit):
        for i in range(m):
            x[i] = (1./Z[i, i]) * (b[i] - np.dot(Z[i,:i], x[:i]) 
                                   - np.dot(Z[i,i+1:], x[i+1:])) 
        residual = sl.norm(Z@x - b)
        residuals.append(residual)
        if (residual < tol): break
    return x



